/* FidoNet interface to SQUISH format message areas
 * Consider this to be a 'private' header
 */

#include "msg.h"                          /* Generic message interface */

                                            /* ----- API publics ----- */
MSGAREA *SquishAreaInit(MSGAREA *ma, void *data);       /* Create area */
void SquishAreaDestroy(MSGAREA *ma);                   /* Destory area */
int SquishOpenArea(MSGAREA *ma);                          /* Open area */
int SquishAreaCreate(MSGAREA *ma);            /* Create/overwrite area */
int SquishAreaClose(MSGAREA *ma);                        /* Close area */
int SquishAreaLock(MSGAREA *ma, long timeout);            /* Lock area */
int SquishAreaUnlock(MSGAREA *ma);                      /* Unlock area */
int SquishKillMsg(MSGAREA *ma, MSGNO which);             /* Kill a msg */
MASTATS SquishAreaStat(MSGAREA *ma);                /* Area statistics */
MSGNO SquishMsgOpen(MSG *msg, MSGNO which, int mode);      /* Open msg */
void SquishMsgClose(MSG *msg);                            /* Close msg */
int SquishMsgRead(MSG *msg);                   /* Read msg into memory */
int SquishMsgWrite(MSG * msg);                /* Write msg from memory */

/* Structures specific to SQUISH
 */

typedef struct sq_area  SQMSGAREA;
typedef struct sq_msg   SQMSG;
typedef struct sq_xmsg  SQXMSG;
typedef struct sq_idx   SQIDX;
typedef struct sq_hdr   SQHDR;
typedef struct sq_base  SQBASE;

#define SQ_FROM_SIZE  36
#define SQ_TO_SIZE    36
#define SQ_SUBJ_SIZE  72
#define SQ_MAX_REPLY  9

struct sq_xmsg
{
  unsigned long attr;
  unsigned char from[SQ_FROM_SIZE];
  unsigned char to[SQ_TO_SIZE];
  unsigned char subj[SQ_SUBJ_SIZE];
  ZNNP          orig;
  ZNNP          dest;
  DATESTAMP     date_written;
  DATESTAMP     date_arrived;
  unsigned short utc_ofs;
  MSGNO         replyto;
  MSGNO         replies[SQ_MAX_REPLY];
  MSGNO         msgno;
  char          ascii_date[20];
};


struct _sqhdr
{
  #define SQHDRID       0xafae4453L
  unsigned long     id;           /* sqhdr.id must always equal SQHDRID */
  long              next_frame;        /* Next frame in the linked list */
  long              prev_frame;       /* Prior frame in the linked list */
  unsigned long     frame_length;               /* Length of this frame */
  unsigned long     msg_length;   /* Length used by XMSG, ctrl and text */
  unsigned long     clen;                   /* Length of ctrl info only */
  #define SQFRAME_NORMAL  0x00  /* Normal text frame */
  #define SQFRAME_FREE    0x01  /* Part of the free chain */
  #define SQFRAME_LZSS    0x02  /* Not implemented */
  #define SQFRAME_UPDATE  0x03  /* Frame is being updated */
  unsigned short    frame_type;        /* Type of frame -- SQFRAME_XXXX */
  unsigned short    filler;
};

struct sq_idx
{
  long              ofs;    /* Offset of the frame relating to this msg */
  MSGNO             msgid;     /* Unique message identifier of this msg */
  #define SQIDXE_MSGREAD  0x80000000Lu  /* Msg read if this is set */
  unsigned long     hash;          /* SquishHash of msg.to for this msg */
};

struct sq_base
{
  unsigned short  len;                 /* LENGTH OF THIS STRUCTURE!   0 */
  unsigned short  filler;                               /* reserved   2 */
  unsigned long   num_msg;            /* Number of messages in area   4 */
  MSGNO           high_msg; /* Highest msg in area. Same as num_msg   8 */
  unsigned long   skip_msg;    /* Skip killing first x msgs in area  12 */
  MSGNO           high_water;                         /* Msg of HWM  16 */
  unsigned long   uid;           /* Number of the next MSGNO to use  20 */
  char            base[80];              /* Base name of SquishFile  24 */
  long            begin_frame;             /* Offset of first frame 104 */
  long            last_frame;               /* Offset to last frame 108 */
  long            free_frame;         /* Offset of first FREE frame 112 */
  long            last_free_frame;     /* Offset of last free frame 116 */
  long            end_frame;              /* Pointer to end of file 120 */
  unsigned long   max_msg;         /* Max # of msgs to keep in area 124 */
  unsigned short  keep_days;             /* Max age of msgs in area 128 */
  unsigned short  sz_sqhdr;                        /* sizeof(SQHDR) 130 */
  char            filler2[124];/* Reserved by Squish for future use 132 */
};                                                        /* total: 256 */

