/* Implements a file with a CRC-16 accumulator
 */

#include <stdarg.h>
#include <string.h>
#include "cfile.h"
#include "mem.h"

CFILE *
cfopen(char const * filename, char const *modes)
{
  CFILE * cfp;
  FILE * fp = fopen(filename, modes);
  if (fp == NULL)
    return NULL;
  cfp = zmalloc("cfopen", sizeof(CFILE));
  cfp->fp = fp;
  cfp->crc16 = 0;
  return cfp;
}

int
cfclose(CFILE * cfp)
{
  int rc = 0;
  if (cfp)
  {
    if (cfp->fp)
      rc = fclose(cfp->fp);
    zfree("cfclose", cfp);
  }
  return rc;
}


size_t
cfwrite(const void *ptr, size_t size, size_t n, CFILE *cfp)
{
  cfp->crc16 = update_crc16(ptr, size*n, cfp->crc16);
  return fwrite(ptr, size, n, cfp->fp);
}

int
cfseek(CFILE *cfp, long int offset, int whence)
{
  return fseek(cfp->fp, offset, whence);
}

char *
cfgets(char *s, int n, CFILE *cfp)
{
  return fgets(s, n, cfp->fp);
}

int
cfprintf(CFILE *cfp, const char *fmt, ...)
{
  int sz;
  va_list argp;
  static char buf[2048];

  va_start(argp, fmt);
  sz = vsprintf(buf, fmt, argp);
  va_end(argp);
  return cfwrite(buf, 1, sz, cfp);
}

int
cfputs(const char *s, CFILE *cfp)
{
  return cfwrite(s, 1, strlen(s), cfp);
}

int
cferror(CFILE * cfp)
{
  return ferror(cfp->fp);
}

