/* osdep.c
 * Operating system dependant functions
 * These are mainly filesystem dependancies
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "osdep.h"
#include "mem.h"
#include "parse.h"

#if defined(DOSISH)
#include <sys/utime.h>
#else
#include <utime.h>
#endif

static char local_path_buf[_MAX_PATH];

#ifdef DOSISH

#include <process.h>
#include <io.h>

#define DIRCH   '\\'
#if !defined(__EMX__)
#include <direct.h>
#endif

/* "portable" directory functions
 */

#if defined(__OS2__) || defined(__EMX__)
  #define INCL_DOS
  #include "os2.h"
  #undef TRUE
  #undef FALSE
  #undef  _SYSINT
  #if defined(__FLAT__) || defined(__EMX__)
  #undef  _FILEFINDBUF
  #define _FILEFINDBUF _FILEFINDBUF3
  #define _SYSINT ULONG
  #define FINDLVL FIL_STANDARD
  #else
  #define _SYSINT USHORT
  #define FINDLVL 0L
  #endif
  #define ALLFILES  "*"

  struct ffind
  {
    HDIR dh;
    char  *directory;
    char  *pattern;
    struct _FILEFINDBUF f;
  };

  #define ff_name(x)     (x)->f.achName
  #define ff_size(x)     (x)->f.cbFile
  #define ff_attr(x)     (x)->f.attrFile
  #define ff_date(x)     *((DOSFileDate *)(&(x)->f.fdateLastWrite))
  #define ff_yr(x)       *((DOSFileDate *)(&(x)->f.fdateLastWrite)).year
  #define ff_mo(x)       *((DOSFileDate *)(&(x)->f.fdateLastWrite)).month
  #define ff_day(x)      *((DOSFileDate *)(&(x)->f.fdateLastWrite)).day
  #define ff_time(x)     *((DOSFileTime *)(&(x)->f.ftimeLastWrite))
  #define ff_hr(x)       *((DOSFileTime *)(&(x)->f.ftimeLastWrite)).hours
  #define ff_min(x)      *((DOSFileTime *)(&(x)->f.ftimeLastWrite)).mins
  #define ff_tsec(x)     *((DOSFileTime *)(&(x)->f.ftimeLastWrite)).tsecs

#elif defined(_WIN32) || defined(WIN32)

  #include <windows.h>
  #define ALLFILES  "*.*"

  typedef struct ffind
  {
    HANDLE handle;
    char  *directory;
    char  *pattern;
    WIN32_FIND_DATA file;
  };

  #define ff_name(x)     (x)->file.cFileName
  #define ff_size(x)     (x)->file.nFileSizeLow
  #define ff_attr(x)     (x)->file.dwFileAttributes
  WORD ff_date(DOSFileData *f);
  WORD ff_yr(DOSFileData *f);
  WORD ff_mo(DOSFileData *f);
  WORD ff_day(DOSFileData *f);
  WORD ff_time(DOSFileData *f);
  WORD ff_hr(DOSFileData *f);
  WORD ff_min(DOSFileData *f);
  WORD ff_tsec(DOSFileData *f);

#elif defined(__MSDOS__) || defined(MSDOS)

 /* Force byte alignment for this braindead struct */

 #if defined(_MSC_VER) || defined(_QC) || defined(__WATCOMC__)
  #pragma pack(1)
 #elif defined(__ZTC__)
  #pragma ZTC align 1
 #elif defined(__TURBOC__) && (__TURBOC__ > 0x202)
  #pragma option -a-
 #endif
 #ifdef __GNUC__
  #define PAK      __attribute__((packed))
 #else
  #define PAK
 #endif
 #define ALLFILES  "*.*"

 #ifdef __TURBOC__
 #include "dir.h"
 #endif
 #include "dos.h"

  /* Structures to access dates and times
   */

  typedef struct
  {
    unsigned tsecs:   5;
    unsigned mins:    6;
    unsigned hours:   5;
  } DOSFileTime;

  typedef struct
  {
    unsigned day:     5;
    unsigned month:   4;
    unsigned year:    7;
  } DOSFileDate;

 typedef struct ffind
 {
   char              reserved[21];
   char              attrib;
   DOSFileTime       time;
   DOSFileDate       date;
   long              d_size;
   char              d_name[13];
   char             *directory;
   char             *pattern;
 };

 #define ff_name(x)     (x)->d_name
 #define ff_size(x)     (x)->d_size
 #define ff_attr(x)     (x)->attrib
 #define ff_date(x)     (x)->date
 #define ff_yr(x)       (x)->date.year
 #define ff_mo(x)       (x)->date.month
 #define ff_day(x)      (x)->date.day
 #define ff_time(x)     (x)->time
 #define ff_hr(x)       (x)->time.hours
 #define ff_min(x)      (x)->time.mins
 #define ff_sec(x)      (x)->time.secs

 /* set structure alignment back to default */

 #if defined (_MSC_VER) || defined(_QC) || defined(__WATCOMC__)
  #pragma pack()
 #elif defined (__ZTC__)
  #pragma ZTC align
 #elif defined(__TURBOC__) && (__TURBOC__ > 0x202)
  #pragma option -a.
 #endif
#endif

static char *
fixpath(char * path)
{
  char * p = path;
  while ((p = strchr(p, '/')) != NULL)
    *p++ = DIRCH;
  return path;
}

static int
getdisk(void)
{
 #if defined(__OS2__)
  #if defined(__FLAT__) || defined(__EMX__)
    ULONG drive, asdf;

    DosQueryCurrentDisk(&drive, &asdf);
  #else
    unsigned short drive;
    unsigned long dont_care;

    DosQCurDisk(&drive, &dont_care);
  #endif
    return (int)drive + 'A' - 1;
 #elif defined(__MSDOS__)
    union REGS regs;

    regs.h.ah = 0x19;
    intdos(&regs, &regs);
    return (int)regs.h.al + 'A';
 #elif defined(_WIN32) || defined(WIN32)
    char szPath[PATHLEN];

    GetCurrentDirectory(PATHLEN, szPath);
    return toupper(*szPath);
 #endif
}


static int
setdisk(int drive)
{
 #ifdef __OS2__
  #if defined(__FLAT__) || defined(__EMX__)
    return DosSetDefaultDisk(drive - 'A' + 1);
  #else
    return DosSelectDisk(drive - 'A' + 1);  /* in OS/2, drive 1=A, 2=B etc */
  #endif
 #elif defined(__MSDOS__)
    union REGS regs;

    regs.h.ah = 0x0E;     /* dos select disk */
    regs.h.dl = (unsigned char) drive - 'A';  /* 0=A, 1=B etc */
    intdos(&regs, &regs);
    return (getdisk() == drive) ? 0 : -1;
 #elif defined(_WIN32) || defined(WIN32)
    char szPath[10];

    sprintf(szPath, "%c:.", drive+'A');
    return SetCurrentDirectory(szPath) ? 0 : -1;
 #endif
}

static int
mydrive(int drv)
{
  return (drv) ? toupper(drv) : getdisk();
}

static int
getcurddir(int drive, char *directory, int chrs)
{
  #if defined(__OS2__)
    #if defined(__FLAT__) || defined(__EMX__)
  ULONG buflen = chrs - 3;
  drive = mydrive(drive);
  *directory++ = (char)drive;
  *directory++ = ':';
  *directory++ = DIRCH;
  return (DosQueryCurrentDir((ULONG)(drive - 'A' + 1), (PBYTE)directory, &buflen)==0 ? 0 : -1);
    #else
  USHORT buflen = chrs - 3;
  drive = mydrive(drive);
  *directory++ = (char)drive;
  *directory++ = ':';
  *directory++ = DIRCH;
  return (DosQCurDir((drive - 'A' + 1), directory, &buflen)==0 ? 0 : -1);
    #endif
  #elif defined(__MSDOS__)
  union REGS r;
  chrs=chrs;
  drive = mydrive(drive);
  *directory++ = (char)drive;
  *directory++ = ':';
  *directory++ = DIRCH;

  r.h.ah=0x47;
  r.h.dl=(unsigned char)(drive - 'A' + 1);
  #if defined(__FLAT__)
  r.x.esi=FP_OFF(directory);
  int386(0x21, &r, &r);
  #else
  {
    struct SREGS sr;
    sr.ds= FP_SEG(directory);
    r.x.si=FP_OFF(directory);
    int86x(0x21, &r, &r, &sr);
  }
  #endif
  return (r.x.cflag==0 ? 0 : -1);
  #elif defined(_WIN32) || defined(WIN32)
  ULONG buflen=chrs;
  int old_disk, rc;

  drive = mydrive(drive);
  old_disk=getdisk();
  if (old_disk != drive)
    setdisk(drive);
  rc=GetCurrentDirectory(buflen, directory);
  if (old_disk != drive)
    setdisk(old_disk);
  return (rc > 0 ? 0 : -1);
  #endif
}


int
getcurdir(char * buf, int buflen)
{
  return getcurddir(0, buf, buflen);
}


char *
resolve_path(char * dest, char const * src, char const * rel)
{
  int   drv = 0,                          /* Get current drive */
        abspath = 0,                      /* Absolute path (ignore relative) */
        r = 0;
  char  * p;
  char  workp[_MAX_PATH*2];

  char * dst = (dest == NULL) ? (dest = local_path_buf) : dest;

  if (src && *src)
    strncpy(workp, src, _MAX_PATH);
  else *workp = '\0';

  fixpath(workp);

  if ((p = strchr(workp, ':')) != NULL)   /* Looks like a drive specifier? */
  {
    if (p > (workp + 1))                  /* Maybe it's a volume */
      abspath = 1;
    else                                  /* Standard drive in path */
    {
      drv = toupper(*workp);
      strcpy(workp, workp + 2);
    }
  }

  if (workp[0] == DIRCH)
    ++abspath;                            /* It's an absolute path */
  if (!abspath)                           /* It's not an absolute path */
  {
    char  relto[_MAX_PATH];

    if (rel)
    { /* Resolve relative path first */
      if (!*rel || resolve_path(relto, rel, NULL) == NULL)
        rel = 0;
      else if (!drv)
        drv = mydrive(relto[0]);
      else if (relto[0] != drv)
        rel = 0;                          /* Not relative if wrong drive */
    }

    if (!rel)                             /* Relative to current directory */
      r = (getcurddir(mydrive(drv), relto, sizeof relto) == 0) ? 0 : -1;

    if (r == 0)
    {
      int b, a = strlen(relto);

      if (a && relto[a-1] != DIRCH)
      {
        relto[a++] = DIRCH;
        relto[a] = 0;
      }
      for (b = strlen(workp); b >= 0; --b)
        workp[b+a] = workp[b];

      while (--a >= 0)
        workp[a] = relto[a];
    }
  }

  if (r == 0)       /* We should now have a full path to resolve */
  {
    int l;

    for (p = workp ; *p ;)
    {
      if (*p == DIRCH)
      {
        char *m = p++;
        if (*p == '.')                    /* One dot */
        {
          if (!*++p || *p == DIRCH)      /* Collapse */
          {
            strcpy(m, p);
            p = m;
            continue;
          }
          if (*p == '.')                /* Two dots */
          {
            if (!*++p || *p == DIRCH)  /* Fold back */
            {
              char *b = m;
              while (b > workp && (*--b != DIRCH))
                ;
              if (*b != DIRCH || b == m)
                return NULL;
              strcpy(b, p);
              p = b;
              continue;
            }
          }
        }
        continue;
      }
      ++p;
    }

    if ((l=strlen(workp)) > 0 && workp[l-1] == DIRCH)
      workp[--l] = 0;

    if (l == 0 || workp[l-1] == ':')      /* Drive spec alone */
    {
      workp[l++] = DIRCH;
      workp[l] = 0;
    }

    if (workp[1] != ':' && !(workp[0] == DIRCH && workp[1] == DIRCH))
    {
      *dst++ = (char)mydrive(drv);
      *dst++ = ':';
    }

    memcpy(dst, workp, l);
    dst[l] = 0;
  }

  return r == 0 ? dest : NULL;
}


static int
file_findqualify(FFIND * ff)
{
  return wildmatch(ff_name(ff), ff->pattern, 1);
}

FFIND *
file_findfirst(char const * dir, char const * patt)
{
  static char const fn[] = "file_findfirst";
  #if defined(__MSDOS__)
  #elif defined(__OS2__)
  _SYSINT cnt = 1;
  #endif  /* defined(_WIN32) || defined(WIN32) */
  char spec[_MAX_PATH];
  FFIND * ff = zmalloc(fn, sizeof(FFIND));
  build_path(spec, dir, ALLFILES, NULL);
  memset(ff, 0, sizeof(FFIND));
  #if defined(__MSDOS__)
   #if defined (_MSC_VER) || defined(_QC) || defined(__WATCOMC__)
  if (_dos_findfirst(spec, _A_HIDDEN|_A_SYSTEM|_A_SUBDIR|_A_ARCH, (struct find_t *)ff) != 0)
   #else
  if (findfirst(spec, (struct ffblk *)ff, _A_HIDDEN|_A_SYSTEM|_A_SUBDIR|_A_ARCH) != 0)
   #endif
  {
    zfree(fn, ff);
    return NULL;
  }
  #elif defined(__OS2__)
  ff->dh = (HDIR) -1;
  if (DosFindFirst ((PBYTE)spec, &ff->dh, (_SYSINT)FILE_HIDDEN|FILE_SYSTEM|FILE_DIRECTORY|FILE_ARCHIVED,
      &ff->f, (_SYSINT)sizeof(struct _FILEFINDBUF), &cnt, FINDLVL) != 0)
  {
    zfree(fn, ff);
    return NULL;
  }
  #else /* defined(_WIN32) || defined(WIN32) */
  if ((ff->handle = FindFirstFile(spec, &(ff->file))) == INVALID_HANDLE_VALUE)
  {
    zfree(fn, ff);
    return NULL;
  }
  #endif
  /* We must allocate these now */
  ff->directory = zstrdup(fn, dir ? dir : "");
  ff->pattern = zstrdup(fn, patt ? patt : "*");
  if (!file_findqualify(ff) && !file_findnext(ff))
  {
    file_findclose(ff);
    return NULL;
  }
  return ff;
}

int
file_findnext(FFIND * ff)
{
  do
  {
  #ifdef __MSDOS__
   #if defined (_MSC_VER) || defined(_QC) || defined(__WATCOMC__)
    if (_dos_findnext((struct find_t *)ff) != 0)
   #else
    if (findnext((struct ffblk *)ff) != 0)
   #endif
      return 0;
  #elif defined(__OS2__)
    _SYSINT cnt = 1;
    if (DosFindNext (ff->dh, &ff->f, sizeof(struct _FILEFINDBUF), &cnt))
      return 0;
  #else /* defined(_WIN32) || defined(WIN32) */
    if (FindNextFile(ff->handle, &(ff->file)) == 0)
      return 0;
  #endif
  }
  while (!file_findqualify(ff));
  return 1;
}

void
file_findclose(FFIND * ff)
{
  if (ff)
  {
    static char const fn[] = "file_findclose";
  #ifdef __MSDOS__
    /* This might seem weird, but this helps free up directory
     * handles on a Netware server, which otherwise leaves them
     * allocated (and may eventually run out), so we keep doing
     * a findnext() until we get to the end of directory
     */
    while (file_findnext(ff))
      ;
   #if defined (_MSC_VER) || defined(_QC) || defined(__WATCOMC__)
   _dos_findclose((struct find_t *)ff);
   #endif
  #elif defined(__OS2__)
    DosFindClose (ff->dh);
  #else /* defined(_WIN32) || defined(WIN32) */
    FindClose(ff->handle);
  #endif
    zfree(fn, ff->directory);
    zfree(fn, ff->pattern);
    zfree(fn, ff);
  }
}

char const *
file_dir(FFIND * ff)
{
  return ff->directory;
}

char const *
file_name(FFIND * ff)
{
  return ff_name(ff);
}

char const *
file_path(FFIND * ff, char * buf)
{
  return build_path(buf, ff->directory, ff_name(ff), NULL);
}

unsigned
file_attr(FFIND * ff)
{
  return ff_attr(ff);
}

long
file_size(FFIND * ff)
{
  return ff_size(ff);
}

time_t
file_date(FFIND * ff)
{
  ff=ff;
  return -1;
}

int
file_isdir(FFIND * ff)          /* Is file a directory? */
{
  #if defined(__MSDOS__) || defined(_WIN32) || defined(WIN32)
  return !!(file_attr(ff) & _A_SUBDIR);
  #elif defined(__OS2__)
  return !!(file_attr(ff) & FILE_DIRECTORY);
  #elif /*  */
  return !!(0);
  #endif
}

int
file_isexe(FFIND * ff)          /* Is file an executable? */
{
  #if defined(__MSDOS__) || defined(_WIN32) || defined(WIN32) || defined(__OS2__)
  /* We have to do this the hard way, from the name */
  char * p = strrchr(ff_name(ff), '.');
  if (p != NULL && (stricmp(p, ".com")==0 || stricmp(p, ".exe") == 0))
    return 1;
  #elif /*  */
  #endif
  return 0;
}

int
file_isreg(FFIND * ff)        /* Is file a regular file? */
{
  return !(file_isdir(ff));   /* We may need to do something more compelx than this */
}

int
file_isdev(FFIND * ff)        /* Is flie a device? 1=chr 2=block */
{
  ff=ff;
  return 0;   /* Cheat, for the time being (until we need it :-)) */
}

#define MAXTRIES  16

char *
file_tempname(char const * dir, char * buf)
{
  char * res = NULL;
  int tries = 0;

  if (dir == NULL && (dir=getenv("TMP"))==NULL)
    dir=getenv("TEMP");
  while (res == NULL && (tries++ < MAXTRIES))
  {
    char fn[10];

    static unsigned counter = 0;

    if (counter++ == 0)
    {
      srand((unsigned)time(NULL));
      counter = (unsigned)rand();
    }
    #ifdef __MSDOS__
    #define getpid()  42
    #endif
    sprintf(fn, "%04x%04x", (unsigned)getpid(), counter);
    res = resolve_path(buf, fn, dir);
    if (access(res, F_OK) == 0 || errno != ENOENT)
      res = NULL;
  }
  return res;
}

#define MAXARGS 64

int
file_execute(char * cmd)
{
  int rc = -1;
  if ((cmd && *cmd) || (cmd=getenv("COMSPEC")) != NULL)
  {
    static char const fn[] = "file_execute";
    char * cmdline = zstrdup(fn,cmd);
    if (!cmdline)
      errno=ENOMEM;
    else
    {
      int argno = 0;
      char * args[MAXARGS+1];

      args[argno++] = parseword(cmdline);
      while ((args[argno] = (argno < MAXARGS) ? parseword(NULL) : NULL) != NULL)
        argno++;
      rc = spawnvp(P_WAIT, args[0], args);
      zfree(fn,cmdline);
    }
  }
  return rc;
}

#else

#define DIRCH   '/'
#include <unistd.h>
#include <dirent.h>

#if defined(POSIXDIR)

  struct ffind
  {
    DIR * dir;
    struct dirent * ent;
    char  *directory;
    char  *pattern;
    int   gotstat;
    struct stat st;
  };

  #define ff_name(x)     (x)->ent->d_name
  #define ff_size(x)     (x)->st.st_size

#define fixpath(x)  (x)

static int
getcurddir(char *directory, int chrs)
{
  char * ptr = getcwd(directory, chrs);
  return ptr == NULL ? (*directory='\0', 0) : strlen(directory);
}

int
getcurdir(char * buf, int buflen)
{
  return getcurddir(buf, buflen);
}


char *
resolve_path(char * dest, char const * src, char const * rel)
{
  int   r = 0;
  char  * p;
  char  workp[_MAX_PATH*2];

  char * dst = (dest == NULL) ? (dest = local_path_buf) : dest;

  if (src && *src)
    strncpy(workp, src, _MAX_PATH);
  else *workp = '\0';

  if (workp[0] != DIRCH)	       /* It's not an absolute path */
  {
    char  relto[_MAX_PATH];

    if (rel && (!*rel || resolve_path(relto, rel, NULL)==NULL))
    { /* Resolve relative path first */
        rel = NULL;
    }

    if (rel == NULL)
      getcurdir(relto, _MAX_PATH);

    if (r == 0)
    {
      int b, a = strlen(relto);

      if (a && relto[a-1] != DIRCH)
      {
        relto[a++] = DIRCH;
        relto[a] = 0;
      }
      for (b = strlen(workp); b >= 0; --b)
        workp[b+a] = workp[b];

      while (--a >= 0)
        workp[a] = relto[a];
    }
  }

  if (r == 0)       /* We should now have a full path to resolve */
  {
    int l;

    for (p = workp ; *p ;)
    {
      if (*p == DIRCH)
      {
        char *m = p++;
        if (*p == '.')                    /* One dot */
        {
          if (!*++p || *p == DIRCH)      /* Collapse */
          {
            strcpy(m, p);
            p = m;
            continue;
          }
          if (*p == '.')                /* Two dots */
          {
            if (!*++p || *p == DIRCH)  /* Fold back */
            {
              char *b = m;
              while (b > workp && (*--b != DIRCH))
                ;
              if (*b != DIRCH || b == m)
                return NULL;
              strcpy(b, p);
              p = b;
              continue;
            }
          }
        }
        continue;
      }
      ++p;
    }

    if ((l=strlen(workp)) > 0 && workp[l-1] == DIRCH)
      workp[--l] = 0;

    memcpy(dst, workp, l);
    dst[l] = 0;
  }

  return r == 0 ? dest : NULL;
}


static int
file_findqualify(FFIND * ff)
{
  return wildmatch(ff_name(ff), ff->pattern, 1);
}

FFIND *
file_findfirst(char const * dir, char const * patt)
{
  static char const fn[] = "file_findfirst";
  FFIND * ff = zmalloc(fn, sizeof(FFIND));
  memset(ff, 0, sizeof(FFIND));
  if ((ff->dir = opendir(dir)) == NULL || (ff->ent = readdir(ff->dir)) == NULL)
  {
    if (ff->dir)
      closedir(ff->dir);
    zfree(fn, ff);
    return NULL;
  }
  /* We must allocate these now */
  ff->directory = zstrdup(fn, dir ? dir : "");
  ff->pattern = zstrdup(fn, patt ? patt : "*");
  if (!file_findqualify(ff) && !file_findnext(ff))
  {
    file_findclose(ff);
    return NULL;
  }
  return ff;
}

int
file_findnext(FFIND * ff)
{
  do
  {
    if ((ff->ent = readdir(ff->dir)) == NULL)
      return 0;
    memset(&ff->st, 0, sizeof(ff->st));
  }
  while (!file_findqualify(ff));
  return 1;
}

void
file_findclose(FFIND * ff)
{
  if (ff)
  {
    static char const fn[] = "file_findclose";
    if (ff->dir)
      closedir(ff->dir);
    zfree(fn, ff->directory);
    zfree(fn, ff->pattern);
    zfree(fn, ff);
  }
}

char const *
file_dir(FFIND * ff)
{
  return ff->directory;
}

char const *
file_name(FFIND * ff)
{
  return ff_name(ff);
}

char const *
file_path(FFIND * ff, char * buf)
{
  return ff->ent ? build_path(buf, ff->directory, ff_name(ff), NULL) : NULL;
}

static int
getfileinfo(FFIND * ff)
{
  if (ff->st.st_dev == 0)
  {
    char buf[_MAX_PATH];

    if (file_path(ff, buf) == NULL || stat(buf, &ff->st) == -1)
      return 0;
  }
  return 1;
}

unsigned
file_attr(FFIND * ff)
{
  return getfileinfo(ff) ? ff->st.st_mode : 0;
}

long
file_size(FFIND * ff)
{
  return getfileinfo(ff) ? ff_size(ff) : -1L;
}

time_t
file_date(FFIND * ff)
{
  return getfileinfo(ff) ? ff->st.st_mtime : 0L;
}

int
file_isdir(FFIND * ff)          /* Is file a directory? */
{
  return getfileinfo(ff) ? S_ISDIR(ff->st.st_mode) : 0;
}

int
file_isexe(FFIND * ff)          /* Is file an executable? */
{
  return getfileinfo(ff) ? !!(ff->st.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)) : 0;
}

int
file_isreg(FFIND * ff)        /* Is file a regular file? */
{
  return getfileinfo(ff) ? S_ISREG(ff->st.st_mode) : 0;
}

int
file_isdev(FFIND * ff)        /* Is flie a device? 1=chr 2=block */
{
  return getfileinfo(ff) ? (S_ISCHR(ff->st.st_mode) || S_ISBLK(ff->st.st_mode)) : 0;
}

/* Usually present */
extern char * tempnam(char const *dir, char const *pfx);

char *
file_tempname(char const * dir, char * buf)
{
  char * res = NULL;
 
  if (dir == NULL && (dir=getenv("TMP"))==NULL && (dir=getenv("TMPDIR"))==NULL)
    dir="/tmp";
  if ((res = tempnam(dir, "TfNx")) != NULL)
  {
    strcpy(buf, res);
    free(res);
  }
  return res;
}

#include <sys/wait.h>
#define MAXARGS 64

int
file_execute(char * cmd)
{
  int rc = -1;
  if ((cmd && *cmd) || (cmd=getenv("SHELL")) != NULL)
  {
    static char const fn[] = "file_execute";
    char * cmdline = zstrdup(fn,cmd);
    if (!cmdline)
      errno=ENOMEM;
    else
    {
      pid_t pid;
      int argno = 0;
      char * args[MAXARGS+1];

      args[argno++] = parseword(cmdline);
      while ((args[argno] = (argno < MAXARGS) ? parseword(NULL) : NULL) != NULL)
        argno++;
      if ((pid = fork()) == 0)	       /* Child process */
      {
	execvp(args[0], args);
	exit(-1);
      }
      if (pid == -1)
	rc = -1;
      else
      {				       /* Wait for child to finish */
	waitpid(pid, &rc, 0);
	rc = WEXITSTATUS(rc);
      }
      zfree(fn,cmdline);
    }
  }
  return rc;
}

#else
#error No directory search implementation!
#endif
#endif

char *
build_path(char * dest, char const * path, ...)
{
  char * p;
  va_list argp;

  if (dest == NULL)
    dest = local_path_buf;
  p = dest;
  va_start(argp, path);
  *p = '\0';
  while (path != NULL)
  {
    if (p > dest)
      if (*(p-1) != DIRCH)
        *p++ = DIRCH;
    strcpy(p, path);
    p += strlen(fixpath(p));
    path = va_arg(argp, char const *);
  }
  va_end(argp);
  return dest;
}


char *
last_component(char const * path)
{
  char * p = strrchr(fixpath((char *)path), DIRCH);
#ifdef DOSISH
  if (p==NULL && (p = strrchr(path, ':'))==NULL)
#else
  if (p == NULL)
#endif
    p = (char *)path;
  else ++p;
  return p;
}

char *
get_path(char * dest, char const * path)
{
  int l = last_component((char *)path) - path;
  if (dest == NULL)
    dest = local_path_buf;
  strncpy(dest, path, l);
  dest[l ? l - 1 : 0] = '\0';
  return dest;
}

char *
path_ext(char * path)
{
  char * last = last_component(path);
  char * p = strrchr(last, '.');
  if (p == NULL)
    p = last + strlen(last);
  return p;
}

char *
add_ext(char * path, char const * ext)
{
  char * eptr = path_ext(path);
  if (*ext != '.')
    *eptr++ = '.';
  strcpy(eptr, ext);
  return path;
}

char *
add_extn(char * path, unsigned numbr)
{
  char eext[16];
  sprintf(eext, "%03u", numbr);
  return add_ext(path, eext);
}

int
setfiletimes(char const * dest, struct stat * st)
{
  struct utimbuf u;
  u.actime  = st->st_atime;
  u.modtime = st->st_mtime;
  return utime(dest, &u);
}

int
makedir(char const * s)
{
#if defined(DOSISH) && !defined(__EMX__)
  return mkdir(s);
#else
  return mkdir(s, 0666);
#endif
}

int
makesubdir(char const * s)
{
  char *p, tmp[_MAX_PATH];
  int rc = 0;

  resolve_path(tmp, s, NULL);
  p = tmp;
  do
  {
    char ch = 0;
    struct stat st;
    char * q = strchr(p, DIRCH);
    if (q && (q == tmp || *(q-1) == ':'))
      q = strchr(q + 1, DIRCH);
    if ((p = q) != NULL)
    {
      ch = *q;
      *q = '\0';
      ++p;
    }
    if (stat(tmp, &st)==0)    /* It exists */
      rc = (S_ISDIR(st.st_mode)) ? 0 : -1;
    else rc = makedir(tmp);
    if (q)
      *q = ch;
  } while (rc == 0 && p != NULL);
  return rc;
}

int
cleandir(char const * dir)
{
  int rc = 0;
  FFIND * ff = file_findfirst(dir, NULL);
  if (ff)
  {
    do
      rc = file_isreg(ff) ? remove(file_path(ff, NULL)) : 0;
    while (rc == 0 && file_findnext(ff));
    file_findclose(ff);
  }
  return rc;
}

int
removedir(char const * dir)
{
  return rmdir(dir);
}

int
changedir(char const * dir)
{
  int rc = chdir(dir);
#ifdef DOSISH
  if (rc == 0 && dir[1] == ':')
    setdisk(toupper(*dir));
#endif
  return rc;
}

int
fmove(char const * dest, char const * src)
{
  return fcopy(dest, src, 1);
}

int
fcopy(char const * dest, char const * src, int moveit)
{
  #ifndef O_BINARY
  #define O_BINARY  0
  #endif
  static char const fn[] = "fcopy";
  char * dpath = get_path(NULL, dest);
  int rc = makesubdir(dpath);
  struct stat st;
  int sfd, dfd;
  int iserrno = 0;

  if (rc != 0 || stat(src, &st) != 0 || (sfd = open(src, O_RDONLY|O_BINARY)) == -1)
    iserrno = errno;
  else
  {
    rc = -1;
    if ((dfd = open(dest, O_RDWR|O_CREAT|O_TRUNC|O_BINARY), S_IREAD|S_IWRITE) == -1)
      iserrno = errno;
    else
    {
      #define XFERBUFSZ   4096
      char * buf = zmalloc(fn, XFERBUFSZ);
      if (buf != NULL)
      {
        int rdcount;

        for (rc = 0 ; rc == 0 && (rdcount = read(sfd, buf, XFERBUFSZ)) != 0 ; )
        {
          int wrcount;

          if ((rdcount == -1) ||
              (wrcount = write(dfd, buf, rdcount)) == -1 ||
              (wrcount != rdcount))       /* Short write */
          {
            rc = -1;
            iserrno = errno;
          }
        }
        zfree(fn, buf);
      }
      if (close(dfd) == EOF && rc == 0)
      {
        iserrno = errno;
        rc = -1;
      }
    }
    close(sfd);
    if (rc != 0)
      remove(dest);
    else if (moveit)
    {
      remove(src);
      setfiletimes(dest, &st);
    }
  }
  errno = iserrno;
  return rc;
}


int
escape(register int c)
{
  switch (c)
  {
    case 't': c = '\t'; break;
    case 'r': c = '\r'; break;
    case 'n': c = '\n'; break;
    case 'a': c = '\a'; break;
    case 'b': c = '\b'; break;
    case 'f': c = '\f'; break;
    case 'v': c = '\v'; break;
  }
  return c;
}

int
wildmatch(register char const * s, register char const * p, int icase)
{
  for (; *p; ++s, ++p)
  {
    register int a = *p;
    switch (a)
    {
      case '#':                               /* Numeric */
        if (!*s || !isdigit(*s))
          return 0;
        break;
      case '?':                               /* Any character */
        if (!*s)
          return 0;
        break;
      case '*':                               /* Any series of characters */
        if (*++p)
        {
          while (wildmatch(s, p, icase) == 0)
            if (*++s == '\0')
              return 0;
        }
      return 1;
      case '[':                               /* Character class */
        {
          register int last = 0400;
          register int matched = 0;
          register int reverse = 0;
          register int s1 = icase ? tolower(*s) : *s;
          if (p[1] == '^')
          {
            ++reverse;
            if (!*++p)
              break;
          }
          for (; *++p && *p != ']'; last = icase ? tolower(*p) : *p)
          {
            a = icase ? tolower(*p) : *p;
            if (a != '-')
            {
              if (a == '\\')
              {
                a = escape(*++p);
                if (icase)
                  a = tolower(a);
              }
              if (a == s1)      /* Character match */
                matched = 1;
            }
            else
            { /* Is a range */
              a = *++p;         /* Look for end of range */
              if (icase)
                a = tolower(a);
              if ((s1 <= a) && (s1 >= last))  /* is it in range? */
                matched = 1;
            }
          }
          if (matched == reverse)
            return 0;
        }
        break;
      case '\\':            /* Escaped character */
        if (!*++p)
          break;
        a = escape(*p);
        /* Fallthru */
      default:                                /* Literal match */
        if (icase)
        {
          if (tolower (*s) == tolower (a))
            break;
        }
        else
        {
          if (*s == a)
            break;
        }
        return 0;
      }
      if (!*p)
        break;
  }
  return (*s == '\0') ? 1 : 0;
}


char *
strlwr(char * buf)
{
  char * p = buf;
  while (*p)
    *p = (char)tolower(*p);
  return buf;
}


char *
strupr(char * buf)
{
  char * p = buf;
  while (*p)
    *p = (char)toupper(*p);
  return buf;
}


int
stricmp(char const * s1, char const * s2)
{
  int r;

  do
  {
    r = toupper(*s1) - toupper(*s2);
    s1++; s2++; /* Outside of toupper() to avoid side-effects */
  }
  while (r == 0 && *s1);
  return r;
}

int
strnicmp(char const * s1, char const * s2, size_t n)
{
  int r;
  do
  {
    if (n-- <= 0)
      return 0;
    r = toupper(*s1) - toupper(*s2);
    s1++; s2++; /* Outside of toupper() to avoid side-effects */
  }
  while (r == 0 && *s1);
  return r;
}

